# coding: utf-8
"""api for working with tasks"""

# docs
from typing import List, NamedTuple, Dict, Optional, Callable
import os
import time
from collections import defaultdict, OrderedDict
import json
from supervisely.task.progress import Progress


from supervisely.api.module_api import ApiField, ModuleApiBase, ModuleWithStatus, WaitingTimeExceeded
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from supervisely.io.fs import get_file_name, ensure_base_path, get_file_hash
from supervisely.collection.str_enum import StrEnum
from supervisely._utils import (
    batched,
    take_with_default
)


class TaskApi(ModuleApiBase, ModuleWithStatus):
    """
    API for working with Tasks. :class:`TaskApi<TaskApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # You can connect to API directly
        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Or you can use API from environment
        os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()

        task_id = 121230
        task_info = api.task.get_info_by_id(task_id)
    """
    class RestartPolicy(StrEnum):
        NEVER = 'never'
        ON_ERROR = 'on_error'

    class PluginTaskType(StrEnum):
        TRAIN = 'train'
        INFERENCE = 'inference'
        INFERENCE_RPC = 'inference_rpc'
        SMART_TOOL = 'smarttool'
        CUSTOM = 'custom'

    class Status(StrEnum):
        QUEUED = 'queued'
        CONSUMED = 'consumed'
        STARTED = 'started'
        DEPLOYED = 'deployed'
        ERROR = 'error'
        FINISHED = 'finished'
        TERMINATING = 'terminating'
        STOPPED = 'stopped'

    def __init__(self, api):
        ModuleApiBase.__init__(self, api)
        ModuleWithStatus.__init__(self)

    def get_list(self, workspace_id: int, filters: Optional[List[Dict[str, str]]] = None) -> List[NamedTuple]:
        """
        List of Tasks in the given Workspace.

        :param workspace_id: Workspace ID.
        :type workspace_id: int
        :param filters: List of params to sort output Projects.
        :type filters: List[dict], optional
        :return: List of Tasks with information for the given Workspace.
        :rtype: :class:`List[NamedTuple]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            workspace_id = 23821

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            task_infos = api.task.get_list(workspace_id)

            task_infos_filter = api.task.get_list(23821, filters=[{'field': 'id', 'operator': '=', 'value': 121230}])
            print(task_infos_filter)
            # Output: [
            #     {
            #         "id": 121230,
            #         "type": "clone",
            #         "status": "finished",
            #         "startedAt": "2019-12-19T12:13:09.702Z",
            #         "finishedAt": "2019-12-19T12:13:09.701Z",
            #         "meta": {
            #             "input": {
            #                 "model": {
            #                     "id": 1849
            #                 },
            #                 "isExternal": true,
            #                 "pluginVersionId": 84479
            #             },
            #             "output": {
            #                 "model": {
            #                     "id": 12380
            #                 },
            #                 "pluginVersionId": 84479
            #             }
            #         },
            #         "description": ""
            #     }
            # ]
        """
        return self.get_list_all_pages('tasks.list',
                                       {ApiField.WORKSPACE_ID: workspace_id, ApiField.FILTER: filters or []})

    def get_info_by_id(self, id: int) -> NamedTuple:
        """
        Get Task information by ID.

        :param id: Task ID in Supervisely.
        :type id: int
        :return: Information about Task.
        :rtype: :class:`NamedTuple`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            task_id = 121230

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            task_info = api.task.get_info_by_id(task_id)
            print(task_info)
            # Output: {
            #     "id": 121230,
            #     "workspaceId": 23821,
            #     "description": "",
            #     "type": "clone",
            #     "status": "finished",
            #     "startedAt": "2019-12-19T12:13:09.702Z",
            #     "finishedAt": "2019-12-19T12:13:09.701Z",
            #     "userId": 16154,
            #     "meta": {
            #         "input": {
            #             "model": {
            #                 "id": 1849
            #             },
            #             "isExternal": true,
            #             "pluginVersionId": 84479
            #         },
            #         "output": {
            #             "model": {
            #                 "id": 12380
            #             },
            #             "pluginVersionId": 84479
            #         }
            #     },
            #     "settings": {},
            #     "agentName": null,
            #     "userLogin": "alexxx",
            #     "teamId": 16087,
            #     "agentId": null
            # }
        """
        return self._get_info_by_id(id, 'tasks.info')

    def get_status(self, task_id: int) -> Status:
        """
        Check status of Task by ID.

        :param id: Task ID in Supervisely.
        :type id: int
        :return: Status object
        :rtype: :class:`Status`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            task_id = 121230

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            task_status = api.task.get_status(task_id)
            print(task_status)
            # Output: finished
        """
        status_str = self.get_info_by_id(task_id)[ApiField.STATUS]  # @TODO: convert json to tuple
        return self.Status(status_str)

    def raise_for_status(self, status: Status) -> None:
        """
        Raise error if Task status is ERROR.

        :param status: Status object.
        :type status: Status
        :return: None
        :rtype: :class:`NoneType`
        """
        if status is self.Status.ERROR:
            raise RuntimeError('Task status: ERROR')

    def wait(self, id: int, target_status: Status, wait_attempts: Optional[int]=None,
             wait_attempt_timeout_sec: Optional[int]=None) -> bool:
        """
        Awaiting achievement by given Task of a given status.

        :param id: Task ID in Supervisely.
        :type id: int
        :param target_status: Status object(status of task we expect to destinate).
        :type target_status: Status
        :param wait_attempts: The number of attempts to determine the status of the task that we are waiting for.
        :type wait_attempts: int, optional
        :param wait_attempt_timeout_sec: Number of seconds for intervals between attempts(raise error if waiting time exceeded).
        :type wait_attempt_timeout_sec: int, optional
        :return: True if the desired status is reached, False otherwise
        :rtype: :class:`bool`
        """
        wait_attempts = wait_attempts or self.MAX_WAIT_ATTEMPTS
        effective_wait_timeout = wait_attempt_timeout_sec or self.WAIT_ATTEMPT_TIMEOUT_SEC
        for attempt in range(wait_attempts):
            status = self.get_status(id)
            self.raise_for_status(status)
            if status is target_status:
                return
            time.sleep(effective_wait_timeout)
        raise WaitingTimeExceeded('Waiting time exceeded')

    def upload_dtl_archive(self, task_id: int, archive_path: str, progress_cb: Optional[Callable]=None):
        encoder = MultipartEncoder({'id': str(task_id).encode('utf-8'),
                                    'name': get_file_name(archive_path),
                                    'archive': (
                                        os.path.basename(archive_path), open(archive_path, 'rb'), 'application/x-tar')})

        def callback(monitor_instance):
            read_mb = monitor_instance.bytes_read / 1024.0 / 1024.0
            if progress_cb is not None:
                progress_cb(read_mb)

        monitor = MultipartEncoderMonitor(encoder, callback)
        self._api.post('tasks.upload.dtl_archive', monitor)

    def _deploy_model(self, agent_id, model_id, plugin_id=None, version=None, restart_policy=RestartPolicy.NEVER,
                      settings=None):
        response = self._api.post('tasks.run.deploy', {ApiField.AGENT_ID: agent_id,
                                                       ApiField.MODEL_ID: model_id,
                                                       ApiField.RESTART_POLICY: restart_policy.value,
                                                       ApiField.SETTINGS: settings or {'gpu_device': 0},
                                                       ApiField.PLUGIN_ID: plugin_id,
                                                       ApiField.VERSION: version})
        return response.json()[ApiField.TASK_ID]

    def get_context(self, id: int) -> Dict:
        """
        Get context information by task ID.

        :param id: Task ID in Supervisely.
        :type id: int
        :return: Context information in dict format
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            task_id = 121230

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            context = api.task.get_context(task_id)
            print(context)
            # Output: {
            #     "team": {
            #         "id": 16087,
            #         "name": "alexxx"
            #     },
            #     "workspace": {
            #         "id": 23821,
            #         "name": "my_super_workspace"
            #     }
            # }
        """
        response = self._api.post('GetTaskContext', {ApiField.ID: id})
        return response.json()

    def _convert_json_info(self, info: dict):
        return info

    def run_dtl(self, workspace_id: int, dtl_graph: Dict, agent_id: Optional[int]=None):
        response = self._api.post('tasks.run.dtl', {ApiField.WORKSPACE_ID: workspace_id,
                                                    ApiField.CONFIG: dtl_graph,
                                                    'advanced': {ApiField.AGENT_ID: agent_id}})
        return response.json()[ApiField.TASK_ID]

    def _run_plugin_task(self, task_type, agent_id, plugin_id, version, config, input_projects, input_models,
                         result_name):
        response = self._api.post('tasks.run.plugin', {'taskType': task_type,
                                                       ApiField.AGENT_ID: agent_id,
                                                       ApiField.PLUGIN_ID: plugin_id,
                                                       ApiField.VERSION: version,
                                                       ApiField.CONFIG: config,
                                                       'projects': input_projects,
                                                       'models': input_models,
                                                       ApiField.NAME: result_name})
        return response.json()[ApiField.TASK_ID]

    def run_train(self, agent_id: int, input_project_id: int, input_model_id: int, result_nn_name: str, train_config: Optional[Dict]=None):
        model_info = self._api.model.get_info_by_id(input_model_id)
        return self._run_plugin_task(task_type=TaskApi.PluginTaskType.TRAIN.value,
                                     agent_id=agent_id,
                                     plugin_id=model_info.plugin_id,
                                     version=None,
                                     input_projects=[input_project_id],
                                     input_models=[input_model_id],
                                     result_name=result_nn_name,
                                     config={} if train_config is None else train_config)

    def run_inference(self, agent_id: int, input_project_id: int, input_model_id: int, result_project_name: str,
                      inference_config:Optional[Dict]=None):
        model_info = self._api.model.get_info_by_id(input_model_id)
        return self._run_plugin_task(task_type=TaskApi.PluginTaskType.INFERENCE.value,
                                     agent_id=agent_id,
                                     plugin_id=model_info.plugin_id,
                                     version=None,
                                     input_projects=[input_project_id],
                                     input_models=[input_model_id],
                                     result_name=result_project_name,
                                     config={} if inference_config is None else inference_config)

    def get_training_metrics(self, task_id: int):
        response = self._get_response_by_id(id=task_id, method='tasks.train-metrics', id_field=ApiField.TASK_ID)
        return response.json() if (response is not None) else None

    def deploy_model(self, agent_id: int, model_id: int) -> int:
        task_ids = self._api.model.get_deploy_tasks(model_id)
        if len(task_ids) == 0:
            task_id = self._deploy_model(agent_id, model_id)
        else:
            task_id = task_ids[0]
        self.wait(task_id, self.Status.DEPLOYED)
        return task_id

    def deploy_model_async(self, agent_id: int, model_id: int) -> int:
        task_ids = self._api.model.get_deploy_tasks(model_id)
        if len(task_ids) == 0:
            task_id = self._deploy_model(agent_id, model_id)
        else:
            task_id = task_ids[0]
        return task_id

    def start(self, agent_id, app_id, workspace_id, description='application description', params=None,
              log_level='info',
              users_ids=None, app_version='', is_branch=False,
              task_name='pythonSpawned', restart_policy='never', proxy_keep_url=False):

        resp = self._api.post(method='tasks.run.app', data={
            ApiField.AGENT_ID: agent_id,
            ApiField.APP_ID: app_id,
            ApiField.WORKSPACE_ID: workspace_id,
            ApiField.DESCRIPTION: description,
            ApiField.PARAMS: take_with_default(params, {'state': {}}),
            ApiField.LOG_LEVEL: log_level,
            ApiField.USERS_IDS: take_with_default(users_ids, []),
            ApiField.APP_VERSION: app_version,
            ApiField.IS_BRANCH: is_branch,
            ApiField.TASK_NAME: task_name,
            ApiField.RESTART_POLICY: restart_policy,
            ApiField.PROXY_KEEP_URL: proxy_keep_url
        })

        return resp.json()

    def stop(self, id):
        response = self._api.post('tasks.stop', {ApiField.ID: id})
        return self.Status(response.json()[ApiField.STATUS])

    def get_import_files_list(self, id: int) -> Dict or None:
        response = self._api.post('tasks.import.files_list', {ApiField.ID: id})
        return response.json() if (response is not None) else None

    def download_import_file(self, id, file_path, save_path):
        response = self._api.post('tasks.import.download_file', {ApiField.ID: id, ApiField.FILENAME: file_path},
                                  stream=True)

        ensure_base_path(save_path)
        with open(save_path, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)

    def create_task_detached(self, workspace_id: int, task_type: Optional[str]=None):
        response = self._api.post('tasks.run.python', {ApiField.WORKSPACE_ID: workspace_id,
                                                       ApiField.SCRIPT: "xxx",
                                                       ApiField.ADVANCED: {ApiField.IGNORE_AGENT: True}})
        return response.json()[ApiField.TASK_ID]

    def submit_logs(self, logs) -> None:
        response = self._api.post('tasks.logs.add', {ApiField.LOGS: logs})
        # return response.json()[ApiField.TASK_ID]

    def upload_files(self, task_id: int, abs_paths: List[str], names: List[str], progress_cb:Optional[Callable] = None) -> None:
        if len(abs_paths) != len(names):
            raise RuntimeError("Inconsistency: len(abs_paths) != len(names)")

        hashes = []
        if len(abs_paths) == 0:
            return

        hash_to_items = defaultdict(list)
        hash_to_name = defaultdict(list)
        for idx, item in enumerate(zip(abs_paths, names)):
            path, name = item
            item_hash = get_file_hash(path)
            hashes.append(item_hash)
            hash_to_items[item_hash].append(path)
            hash_to_name[item_hash].append(name)

        unique_hashes = set(hashes)
        remote_hashes = self._api.image.check_existing_hashes(list(unique_hashes))
        new_hashes = unique_hashes - set(remote_hashes)

        # @TODO: upload remote hashes
        if len(remote_hashes) != 0:
            files = []
            for hash in remote_hashes:
                for name in hash_to_name[hash]:
                    files.append({ApiField.NAME: name, ApiField.HASH: hash})
            for batch in batched(files):
                resp = self._api.post('tasks.files.bulk.add-by-hash',
                                      {ApiField.TASK_ID: task_id, ApiField.FILES: batch})
        if progress_cb is not None:
            progress_cb(len(remote_hashes))

        for batch in batched(list(zip(abs_paths, names, hashes))):
            content_dict = OrderedDict()
            for idx, item in enumerate(batch):
                path, name, hash = item
                if hash in remote_hashes:
                    continue
                content_dict["{}".format(idx)] = json.dumps({"fullpath": name, "hash": hash})
                content_dict["{}-file".format(idx)] = (name, open(path, 'rb'), '')

            if len(content_dict) > 0:
                encoder = MultipartEncoder(fields=content_dict)
                resp = self._api.post('tasks.files.bulk.upload', encoder)
                if progress_cb is not None:
                    progress_cb(len(content_dict))

    # {
    #     data: {my_val: 1}
    #     obj: {val: 1, res: 2}
    # }
    # {
    #     obj: {new_val: 1}
    # }
    # // apped: true, recursive: false
    # {
    #     data: {my_val: 1}
    #     obj: {new_val: 1}
    # }(edited)
    # // append: false, recursive: false
    # {
    #     obj: {new_val: 1}
    # }(edited)
    #
    # 16: 32
    # // append: true, recursive: true
    # {
    #     data: {my_val: 1}
    #     obj: {val: 1, res: 2, new_val: 1}
    # }

    def set_fields(self, task_id: int, fields: List) -> Dict:
        for idx, obj in enumerate(fields):
            for key in [ApiField.FIELD, ApiField.PAYLOAD]:
                if key not in obj:
                    raise KeyError("Object #{} does not have field {!r}".format(idx, key))
        data = {
            ApiField.TASK_ID: task_id,
            ApiField.FIELDS: fields
        }
        resp = self._api.post('tasks.data.set', data)
        return resp.json()

    def set_fields_from_dict(self, task_id: int, d: Dict) -> Dict:
        fields = []
        for k, v in d.items():
            fields.append({ApiField.FIELD: k, ApiField.PAYLOAD: v})
        return self.set_fields(task_id, fields)

    def set_field(self, task_id: int, field: Dict, payload: Dict, append: Optional[bool]=False, recursive: Optional[bool]=False) -> Dict:
        fields = [
            {
                ApiField.FIELD: field,
                ApiField.PAYLOAD: payload,
                ApiField.APPEND: append,
                ApiField.RECURSIVE: recursive,
            }
        ]
        return self.set_fields(task_id, fields)

    def get_fields(self, task_id, fields: List):
        data = {
            ApiField.TASK_ID: task_id,
            ApiField.FIELDS: fields
        }
        resp = self._api.post('tasks.data.get', data)
        return resp.json()["result"]

    def get_field(self, task_id: int, field: Dict):
        result = self.get_fields(task_id, [field])
        return result[field]

    def _validate_checkpoints_support(self, task_id):
        info = self.get_info_by_id(task_id)
        if info["type"] != str(TaskApi.PluginTaskType.TRAIN):
            raise RuntimeError("Task (id={!r}) has type {!r}. "
                               "Checkpoints are available only for tasks of type {!r}".format())

    def list_checkpoints(self, task_id: int):
        self._validate_checkpoints_support(task_id)
        resp = self._api.post('tasks.checkpoints.list', {ApiField.ID: task_id})
        return resp.json()

    def delete_unused_checkpoints(self, task_id: int) -> Dict:
        self._validate_checkpoints_support(task_id)
        resp = self._api.post("tasks.checkpoints.clear", {ApiField.ID: task_id})
        return resp.json()

    def _set_output(self):
        pass

    def set_output_project(self, task_id: int, project_id: int, project_name: Optional[str]=None) -> Dict:
        if project_name is None:
            project = self._api.project.get_info_by_id(project_id)
            project_name = project.name

        output = {
            ApiField.PROJECT: {
                ApiField.ID: project_id,
                ApiField.TITLE: project_name
            }
        }
        resp = self._api.post("tasks.output.set", {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output})
        return resp.json()

    def set_output_report(self, task_id: int, file_id: int, file_name: str) -> Dict:
        return self._set_custom_output(task_id, file_id, file_name, description="Report", icon="zmdi zmdi-receipt")

    def _set_custom_output(self, task_id, file_id, file_name, file_url=None, description="File",
                           icon="zmdi zmdi-file-text", color="#33c94c", background_color="#d9f7e4",
                           download=False):
        if file_url is None:
            file_url = self._api.file.get_url(file_id)

        output = {
            ApiField.GENERAL: {
                "icon": {
                    "className": icon,
                    "color": color,
                    "backgroundColor": background_color
                },
                "title": file_name,
                "titleUrl": file_url,
                "download": download,
                "description": description
            }
        }
        resp = self._api.post("tasks.output.set", {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output})
        return resp.json()

    def set_output_archive(self, task_id: int, file_id: int, file_name: str, file_url: Optional[str]=None) -> Dict:
        if file_url is None:
            file_url = self._api.file.get_info_by_id(file_id).storage_path
        return self._set_custom_output(task_id, file_id, file_name,
                                       file_url=file_url,
                                       description="Download archive", icon="zmdi zmdi-archive",
                                       download=True)

    def set_output_file_download(self, task_id: int, file_id: int, file_name: str, file_url: Optional[str]=None,
                                 download: Optional[bool]=True) -> Dict:
        if file_url is None:
            file_url = self._api.file.get_info_by_id(file_id).storage_path
        return self._set_custom_output(task_id, file_id, file_name,
                                       file_url=file_url,
                                       description="Download file", icon="zmdi zmdi-file",
                                       download=download)

    def send_request(self, task_id: int, method: str, data: Dict, context: Optional[Dict]={},
                     skip_response: bool=False, timeout: Optional[int]=60):
        if type(data) is not dict:
            raise TypeError("data argument has to be a dict")
        resp = self._api.post("tasks.request.direct", {ApiField.TASK_ID: task_id,
                                                       ApiField.COMMAND: method,
                                                       ApiField.CONTEXT: context,
                                                       ApiField.STATE: data,
                                                       'skipResponse': skip_response,
                                                       'timeout': timeout})
        return resp.json()

    def set_output_directory(self, task_id, file_id, directory_path):
        return self._set_custom_output(task_id, file_id, directory_path, description="Directory",
                                       icon="zmdi zmdi-folder")

    def update_meta(self, id: int, data: dict):
        """
        Update given task metadata
        :param id: int — task id
        :param data: dict — meta data to update

        """
        if type(data) == dict:
            data.update({'id': id})

        self._api.post("tasks.meta.update", data)



#!/usr/bin/env python
# coding: utf-8

"""A module for reading NRRD files [NRRD1]_, basically a wrapper for calls on the pynrrd library [NRRD2]_.

References
----------
.. [NRRD1] http://teem.sourceforge.net/nrrd/format.html (20180212)
.. [NRRD2] https://github.com/mhe/pynrrd (20180212).
"""

import nrrd
import numpy as np

from loaders.volume import Volume


def open_image(path, verbose=True):
    """
    Open a 3D NRRD image at the given path.

    Parameters
    ----------
    path : str
        The path of the file to be loaded.
    verbose : bool, optional
        If `True` (default), print some meta data of the loaded file to standard output.

    Returns
    -------
    Volume
        The resulting 3D image volume, with the ``src_object`` attribute set to the tuple `(data, header)` returned
        by pynrrd's ``nrrd.read`` (where `data` is a Numpy array and `header` is a dictionary) and the desired
        anatomical world coordinate system ``system`` set to "RAS".

    Raises
    ------
    IOError
        If something goes wrong.
    """
    try:
        src_object = (voxel_data, hdr) = nrrd.read(path)
    except Exception as e:
        raise IOError(e)

    if verbose:
        print("Loading image:", path)
        print("Meta data:")
        for k in sorted(hdr.keys(), key=str.lower):
            print("{}: {!r}".format(k, hdr[k]))

    __check_data_kinds_in(hdr)
    src_system = __world_coordinate_system_from(
        hdr
    )  # No fixed world coordinates for NRRD images!
    mat = __matrix_from(hdr)  # Voxels to world coordinates

    # Create new ``Volume`` instance
    volume = Volume(
        src_voxel_data=voxel_data,
        src_transformation=mat,
        src_system=src_system,
        system="RAS",
        src_object=src_object,
    )
    return volume


def save_image(path, data, transformation, system="RAS", kinds=None):
    """
    Save the given image data as a NRRD image file at the given path.

    Parameters
    ----------
    path : str
        The path for the file to be saved.
    data : array_like
        Three-dimensional array that contains the voxels to be saved.
    transformation : array_like
        :math:`4×4` transformation matrix that maps from ``data``'s voxel indices to the given ``system``'s anatomical
        world coordinate system.
    system : str, optional
        The world coordinate system to which ``transformation`` maps the voxel data. Either "RAS" (default), "LAS", or
        "LPS" (these are the ones supported by the NRRD format).
    kinds : str or sequence of strings, optional
        If given, the string(s) will be used to set the NRRD header's "kinds" field. If a single string is given, it
        will be used for all dimensions. If multiple strings are given, they will be used in the given order. If
        nothing is given (default), the "kinds" field will not be set. Note that all strings should either be "domain"
        or "space".

    """
    if data.ndim > 3:
        raise RuntimeError(
            "Currently, only supports saving NRRD files with scalar data only!"
        )

    # Create the header entries from the transformation
    space = system.upper()
    space_directions = transformation[:3, :3].T.tolist()
    space_origin = transformation[:3, 3].tolist()
    options = {
        "space": space,
        "space directions": space_directions,
        "space origin": space_origin,
    }
    if kinds is not None:
        kinds = (data.ndim * [kinds]) if isinstance(kinds, str) else list(kinds)
        options["kinds"] = kinds
    nrrd.write(filename=path, data=data, header=options)


def save_volume(path, volume, src_order=True, src_system=True, kinds=None):
    """
    Save the given ``Volume`` instance as a NRRD image file at the given path.

    Parameters
    ----------
    path : str
        The path for the file to be saved.
    volume : Volume
        The ``Volume`` instance containing the image data to be saved.
    src_order : bool, optional
        If `True` (default), order the saved voxels as in ``src_data``; if `False`, order the saved voxels as in
        ``aligned_data``. In any case, the correct transformation matrix will be chosen.
    src_system : bool, optional
        If `True` (default), try to use ``volume``'s ``src_system`` as the anatomical world coordinate system for
        saving; if `False`, try to use ``volume``'s ``system`` instead. In either case, this works if the system is
        either "RAS", "LAS", or "LPS" (these are the ones supported by the NRRD format). If a different system is
        given, use "RAS" instead.
    kinds : str or sequence of strings, optional
        If given, the string(s) will be used to set the NRRD header's "kinds" field. If a single string is given, it
        will be used for all dimensions. If multiple strings are given, they will be used in the given order. If
        nothing is given (default), the "kinds" field will not be set. Note that all strings should either be "domain"
        or "space".
    """
    if volume.aligned_data.ndim > 3:
        raise RuntimeError(
            "Currently, only supports saving NRRD files with scalar data only!"
        )

    system = volume.src_system if src_system else volume.system
    system = system if system in ["RAS", "LAS", "LPS"] else "RAS"

    if src_order:
        data = volume.src_data
        transformation = volume.get_src_transformation(system)
    else:
        data = volume.aligned_data
        transformation = volume.get_aligned_transformation(system)

    save_image(
        path, data=data, transformation=transformation, system=system, kinds=kinds
    )


def __check_data_kinds_in(header):
    """
    Sanity check on the header's "kinds" field: are all entries either "domain" or "space" (i.e. are we really dealing
    with scalar data on a spatial domain)?

    Parameters
    ----------
    header : dict
        A dictionary containing the NRRD header (as returned by ``nrrd.read``, for example).

    Returns
    -------
    None
        Simply return if everything is ok or the "kinds" field is not set.

    Raises
    ------
    IOError
        If the "kinds" field contains entries other than "domain" or "space".
    """
    kinds = header.get("kinds")
    if kinds is None:
        return

    for k in kinds:
        if k.lower() not in ["domain", "space"]:
            raise IOError("At least one data dimension contains non-spatial data!")


def __world_coordinate_system_from(header):
    """
    From the given NRRD header, determine the respective assumed anatomical world coordinate system.

    Parameters
    ----------
    header : dict
        A dictionary containing the NRRD header (as returned by ``nrrd.read``, for example).

    Returns
    -------
    str
        The three-character uppercase string determining the respective anatomical world coordinate system (such as
        "RAS" or "LPS").

    Raises
    ------
    IOError
        If the header is missing the "space" field or the "space" field's value does not determine an anatomical world
        coordinate system.
    """
    try:
        system_str = header["space"]
    except KeyError as e:
        raise IOError(
            "Need the header's \"space\" field to determine the image's anatomical coordinate system."
        )

    if len(system_str) == 3:
        # We are lucky: this is already the format that we need
        return system_str.upper()

    # We need to separate the string (such as "right-anterior-superior") at its dashes, then get the first character
    # of each component. We cannot handle 4D data nor data with scanner-based coordinates ("scanner-...") or
    # non-anatomical coordinates ("3D-...")
    system_components = system_str.split("-")
    if len(system_components) == 3 and not system_components[0].lower() in [
        "scanner",
        "3d",
    ]:
        system_str = "".join(c[0].upper() for c in system_components)
        return system_str

    raise IOError('Cannot handle "space" value {}'.format(system_str))


def __matrix_from(header):
    """
    Calculate the transformation matrix from voxel coordinates to the header's anatomical world coordinate system.

    Parameters
    ----------
    header : dict
        A dictionary containing the NRRD header (as returned by ``nrrd.read``, for example).

    Returns
    -------
    numpy.ndarray
        The resulting :math:`4×4` transformation matrix.
    """
    try:
        space_directions = header["space directions"]
        space_origin = header["space origin"]
    except KeyError as e:
        raise IOError(
            'Need the header\'s "{}" field to determine the mapping from voxels to world coordinates.'.format(
                e
            )
        )

    # "... the space directions field gives, one column at a time, the mapping from image space to world space
    # coordinates ... [1]_" -> list of columns, needs to be transposed
    trans_3x3 = np.array(space_directions).T
    trans_4x4 = np.eye(4)
    trans_4x4[:3, :3] = trans_3x3
    trans_4x4[:3, 3] = space_origin
    return trans_4x4



# coding: utf-8


# docs
from __future__ import annotations
import cv2
import numpy as np
from typing import List, Dict, Optional
from supervisely.geometry.point_location import PointLocation
from supervisely.geometry.rectangle import Rectangle


from shapely.geometry import mapping, Polygon as ShapelyPolygon

from supervisely.geometry.conversions import shapely_figure_to_coords_list
from supervisely.geometry.point_location import row_col_list_to_points, points_to_row_col_list
from supervisely.geometry.vector_geometry import VectorGeometry
from supervisely.geometry.constants import EXTERIOR, INTERIOR, POINTS, LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID
from supervisely.geometry import validation
from supervisely.sly_logger import logger


class Polygon(VectorGeometry):
    """
    Polygon geometry for a single :class:`Label<supervisely.annotation.label.Label>`. :class:`Polygon<Polygon>` class object is immutable.

    :param exterior: List of :class:`PointLocation<supervisely.geometry.point_location.PointLocation>` objects, the object contour is defined with these points.
    :type exterior: List[PointLocation]
    :param interior: List of :class:`PointLocation<supervisely.geometry.point_location.PointLocation>` objects, the object holes is defined with these points.
    :type interior: List[List[PointLocation]]
    :param sly_id: Polygon ID in Supervisely server.
    :type sly_id: int, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which Polygon belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created Polygon.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when Polygon was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when Polygon was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :raises: :class:`ValueError`, if len(exterior) < 3 or len(any element in interior list) < 3

    :Usage example:

     .. code-block:: python

            import supervisely as sly

            exterior = [sly.PointLocation(730, 2104), sly.PointLocation(2479, 402), sly.PointLocation(3746, 1646)]
            interior = [[sly.PointLocation(1907, 1255), sly.PointLocation(2468, 875), sly.PointLocation(2679, 1577)]]
            figure = sly.Polygon(exterior, interior)
    """
    @staticmethod
    def geometry_name():
        return 'polygon'

    def __init__(self, exterior: List[PointLocation], interior: List[List[PointLocation]],
                 sly_id: Optional[int] = None, class_id: Optional[int] = None, labeler_login: Optional[int] = None,
                 updated_at: Optional[str] = None, created_at: Optional[str] = None):
        if len(exterior) < 3:
            exterior.extend([exterior[-1]] * (3 - len(exterior)))
            logger.warn('"{}" field must contain at least 3 points to create "Polygon" object.'.format(EXTERIOR))
            #raise ValueError('"{}" field must contain at least 3 points to create "Polygon" object.'.format(EXTERIOR))

        for element in interior:
            if len(element) < 3:
                logger.warn('"{}" interior field must contain at least 3 points to create "Polygon" object.'.format(element))
                element.extend([element[-1]] * (3 - len(element)))
        #if any(len(element) < 3 for element in interior):
        #    raise ValueError('"{}" element must contain at least 3 points.'.format(INTERIOR))

        super().__init__(exterior, interior, sly_id=sly_id, class_id=class_id, labeler_login=labeler_login,
                         updated_at=updated_at, created_at=created_at)

    @classmethod
    def from_json(cls, data: Dict) -> Polygon:
        """
        Convert a json dict to Polygon. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: Polygon in json format as a dict.
        :type data: dict
        :return: Polygon object
        :rtype: :class:`Polygon<Polygon>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            figure_json =  {
                "points": {
                    "exterior": [
                        [2104, 730],
                        [402, 2479],
                        [1646, 3746]
                    ],
                    "interior": [
                        [
                            [1255, 1907],
                            [875, 2468],
                            [577, 2679]
                        ]
                    ]
                }
            }

            figure = sly.Polygon.from_json(figure_json)
        """
        validation.validate_geometry_points_fields(data)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(exterior=row_col_list_to_points(data[POINTS][EXTERIOR], flip_row_col_order=True),
                   interior=[row_col_list_to_points(i, flip_row_col_order=True) for i in data[POINTS][INTERIOR]],
                   sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    def crop(self, rect: Rectangle) -> List[Polygon]:
        """
        Crops current Polygon.

        :param rect: Rectangle object for crop.
        :type rect: Rectangle
        :return: List of Polygon objects
        :rtype: :class:`List[Polygon]<Polygon>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            crop_figures = figure.crop(sly.Rectangle(1, 1, 300, 350))
        """
        try:
            # points = [
            #     PointLocation(row=rect.top, col=rect.left),
            #     PointLocation(row=rect.top, col=rect.right + 1),
            #     PointLocation(row=rect.bottom + 1, col=rect.right + 1),
            #     PointLocation(row=rect.bottom + 1, col=rect.left)
            # ]
            points = [
                PointLocation(row=rect.top, col=rect.left),
                PointLocation(row=rect.top, col=rect.right),
                PointLocation(row=rect.bottom, col=rect.right),
                PointLocation(row=rect.bottom, col=rect.left)
            ]
            #points = rect.corners # old implementation with 1 pixel error (right bottom)
            # #@TODO: investigate here (critical issue)

            clipping_window_shpl = ShapelyPolygon(points_to_row_col_list(points))
            self_shpl = ShapelyPolygon(self.exterior_np, holes=self.interior_np)
            intersections_shpl = self_shpl.buffer(0).intersection(clipping_window_shpl)
            mapping_shpl = mapping(intersections_shpl)
        except Exception:
            logger.warn('Polygon cropping exception, shapely.', exc_info=True)
            # raise
            # if polygon is invalid, just print warning and skip it
            # @TODO: need more investigation here
            return []

        intersections = shapely_figure_to_coords_list(mapping_shpl)

        # Check for bad cropping cases (e.g. empty points list)
        out_polygons = []
        for intersection in intersections:
            if isinstance(intersection, list) and len(intersection) > 0 and len(intersection[0]) >= 3:
                exterior = row_col_list_to_points(intersection[0], do_round=True)
                interiors = []
                for interior_contour in intersection[1:]:
                    if len(interior_contour) > 2:
                        interiors.append(row_col_list_to_points(interior_contour, do_round=True))
                out_polygons.append(Polygon(exterior, interiors))
        return out_polygons

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        exterior = self.exterior_np[:, ::-1]
        interior = [x[:, ::-1] for x in self.interior_np]
        bmp_to_draw = np.zeros(bitmap.shape[:2], np.uint8)
        cv2.fillPoly(bmp_to_draw, pts=[exterior], color=1)
        cv2.fillPoly(bmp_to_draw, pts=interior, color=0)
        bool_mask = bmp_to_draw.astype(bool)
        bitmap[bool_mask] = color

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        exterior = self.exterior_np[:, ::-1]
        interior = [x[:, ::-1] for x in self.interior_np]

        poly_lines = [exterior] + interior
        cv2.polylines(bitmap, pts=poly_lines, isClosed=True, color=color, thickness=thickness)

    # @TODO: extend possibilities, consider interior
    # returns area of exterior figure only
    @property
    def area(self) -> float:
        """
        Polygon area.

        :return: Area of current Polygon object (exterior figure only).
        :rtype: :class:`float`

        :Usage Example:

         .. code-block:: python

            print(figure.area)
            # Output: 7288.0
        """
        exterior = self.exterior_np
        return self._get_area_by_gauss_formula(exterior[:, 0], exterior[:, 1])

    @staticmethod
    def _get_area_by_gauss_formula(rows, cols):
        return 0.5 * np.abs(np.dot(rows, np.roll(cols, 1)) - np.dot(cols, np.roll(rows, 1)))

    def approx_dp(self, epsilon: float) -> Polygon:
        """
        Approximates a Polygon curve with the specified precision.

        :param epsilon: Specifying the approximation accuracy.
        :type epsilon: float
        :return: Polygon object
        :rtype: :class:`Polygon<Polygon>`

        :Usage Example:

         .. code-block:: python

            # Remember that Polygon class object is immutable, and we need to assign new instance of Polygon to a new variable
            approx_figure = figure.approx_dp(0.75)
        """
        exterior_np = self._approx_ring_dp(self.exterior_np, epsilon, closed=True).tolist()
        interior_np = [self._approx_ring_dp(x, epsilon, closed=True).tolist() for x in self.interior_np]
        exterior = row_col_list_to_points(exterior_np, do_round=True)
        interior = [row_col_list_to_points(x, do_round=True) for x in interior_np]
        return Polygon(exterior, interior)

    @classmethod
    def allowed_transforms(cls):
        from supervisely.geometry.any_geometry import AnyGeometry
        from supervisely.geometry.rectangle import Rectangle
        from supervisely.geometry.bitmap import Bitmap
        return [AnyGeometry, Rectangle, Bitmap]



# coding: utf-8

# docs
from __future__ import annotations
import numpy as np

from supervisely.api.module_api import ApiField, ModuleApi
from supervisely.io.fs import ensure_base_path
from supervisely.imaging.image import read_bytes


class VideoFrameAPI(ModuleApi):
    """
    :class:`Frame<supervisely.video_annotation.frame.Frame>` for a single video. :class:`VideoFrameAPI<VideoFrameAPI>` object is immutable.
    """
    def _download(self, video_id, frame_index):
        """
        :param video_id: int
        :param frame_index: int
        :return: Response class object containing frame data with given index from given video id
        """
        response = self._api.post('videos.download-frame', {ApiField.VIDEO_ID: video_id, ApiField.FRAME: frame_index})
        return response

    def download_np(self, video_id: int, frame_index: int) -> np.ndarray:
        """
        Download Image for frame with given index from given video ID in numpy format(RGB).

        :param video_id: Video ID in Supervisely.
        :type video_id: int
        :param frame_index: Index of frame to download.
        :type frame_index: int
        :return: Image in RGB numpy matrix format
        :rtype: :class:`np.ndarray`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            video_id = 198703211
            frame_idx = 5
            image_np = api.video.frame.download_np(video_id, frame_idx)
        """
        response = self._download(video_id, frame_index)
        img = read_bytes(response.content)
        return img

    def download_path(self, video_id: int, frame_index: int, path: str) -> None:
        """
        Downloads Image on the given path for frame with given index from given Video ID.

        :param video_id: Video ID in Supervisely.
        :type video_id: int
        :param frame_index: Index of frame to download.
        :type frame_index: int
        :param path: Local save path for Image.
        :type path: str
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            video_id = 198703211
            frame_idx = 5
            save_path = '/home/admin/Downloads/img/result.png'
            api.video.frame.download_path(video_id, frame_idx, save_path)
        """
        response = self._download(video_id, frame_index)
        ensure_base_path(path)
        with open(path, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024*1024):
                fd.write(chunk)



import os
import signal
import psutil
import sys

from fastapi import (
    FastAPI,
    Request,
    Response,
    WebSocket,
    WebSocketDisconnect,
    Depends,
    HTTPException,
)
from fastapi.exception_handlers import http_exception_handler
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles

from supervisely.app.fastapi.templating import Jinja2Templates
from supervisely.app.fastapi.websocket import WebsocketManager
from supervisely.io.fs import mkdir, dir_exists
from supervisely.sly_logger import logger
from supervisely.api.api import SERVER_ADDRESS, API_TOKEN, TASK_ID, Api

# print(supervisely.__path__)
# "--reload-include", "*.py,*.html"


def create() -> FastAPI:
    from supervisely.app import DataJson, StateJson

    app = FastAPI()
    WebsocketManager().set_app(app)

    @app.post("/data")
    async def send_data(request: Request):
        data = DataJson()
        response = JSONResponse(content=dict(data))
        return response

    @app.post("/state")
    async def send_state(request: Request):
        state = StateJson()
        response = JSONResponse(content=dict(state))

        gettrace = getattr(sys, "gettrace", None)
        if gettrace is not None and gettrace():
            response.headers["x-debug-mode"] = "1"
        return response

    @app.post("/session-info")
    async def send_session_info(request: Request):
        server_address = os.environ.get(SERVER_ADDRESS)
        if server_address is not None:
            server_address = Api.normalize_server_address(server_address)

        response = JSONResponse(
            content={
                TASK_ID: os.environ.get(TASK_ID),
                SERVER_ADDRESS: server_address,
                API_TOKEN: os.environ.get(API_TOKEN),
            }
        )
        return response

    @app.post("/shutdown")
    async def shutdown_endpoint(request: Request):
        shutdown()

    @app.websocket("/ws")
    async def websocket_endpoint(websocket: WebSocket):
        await WebsocketManager().connect(websocket)
        try:
            while True:
                data = await websocket.receive_text()
        except WebSocketDisconnect:
            WebsocketManager().disconnect(websocket)

    import supervisely

    app.mount("/css", StaticFiles(directory=supervisely.__path__[0]), name="sly_static")
    return app


def shutdown():
    logger.info("Shutting down...")
    current_process = psutil.Process(os.getpid())
    current_process.send_signal(signal.SIGINT)  # emit ctrl + c


def enable_hot_reload_on_debug(app: FastAPI):
    templates = Jinja2Templates()
    gettrace = getattr(sys, "gettrace", None)
    if gettrace is None:
        print("Can not detect debug mode, no sys.gettrace")
    elif gettrace():
        print("In debug mode ...")
        import arel

        hot_reload = arel.HotReload(paths=[arel.Path(".")])
        app.add_websocket_route("/hot-reload", route=hot_reload, name="hot-reload")
        app.add_event_handler("startup", hot_reload.startup)
        app.add_event_handler("shutdown", hot_reload.shutdown)
        templates.env.globals["DEBUG"] = "1"
        templates.env.globals["hot_reload"] = hot_reload
    else:
        print("In runtime mode ...")


def handle_server_errors(app: FastAPI):
    @app.exception_handler(500)
    async def server_exception_handler(request, exc):
        return await http_exception_handler(
            request,
            HTTPException(
                status_code=500,
                detail={
                    # "title": "error title",
                    "message": repr(exc)
                },
            ),
        )


def init(app: FastAPI, root_dir: str = "."):
    Jinja2Templates(directory=root_dir)
    enable_hot_reload_on_debug(app)
    app.mount("/sly", create())
    handle_server_errors(app)



from supervisely.app.fastapi.subapp import (
    create,
    shutdown,
    enable_hot_reload_on_debug,
    init,
)
from supervisely.app.fastapi.templating import Jinja2Templates
from supervisely.app.fastapi.websocket import WebsocketManager
from supervisely.app.fastapi.utils import run_sync
from supervisely.app.fastapi.offline import available_after_shutdown






from typing import Literal
from supervisely.api.module_api import ApiField, ModuleApiBase


class GithubApi(ModuleApiBase):
    def get_account_info(self):
        response = self._api.post("github.user.info", {})
        return response.json()

    def create_repo(self, name, visibility: Literal["private", "public"] = "private"):
        response = self._api.post(
            "github.user.info", {"repository": name, "visibility": visibility}
        )
        return response.json()



import functools
import os
import pathlib
import shutil
import tempfile
import threading
import traceback
from collections import namedtuple
from distutils.dir_util import copy_tree

from fastapi import FastAPI
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles

import supervisely as sly


def get_static_paths_by_mounted_object(mount) -> list:
    StaticPath = namedtuple('StaticPath', ['local_path', 'url_path'])
    static_paths = []

    if hasattr(mount, 'routes'):
        for current_route in mount.routes:
            if type(current_route) == Mount and type(current_route.app) == FastAPI:
                all_children_paths = get_static_paths_by_mounted_object(current_route)
                for index, current_path in enumerate(all_children_paths):
                    current_url_path = pathlib.Path(str(current_route.path).lstrip('/'),
                                                    str(current_path.url_path).lstrip('/'))
                    all_children_paths[index] = StaticPath(local_path=current_path.local_path,
                                                           url_path=current_url_path)
                static_paths.extend(all_children_paths)
            elif type(current_route) == Mount and type(current_route.app) == StaticFiles:
                static_paths.append(StaticPath(local_path=pathlib.Path(current_route.app.directory),
                                               url_path=pathlib.Path(str(current_route.path).lstrip('/'))))

    return static_paths


def dump_statics_to_dir(static_dir_path: pathlib.Path, static_paths: list):
    for current_path in static_paths:
        current_local_path: pathlib.Path = current_path.local_path
        current_url_path: pathlib.Path = static_dir_path / current_path.url_path

        if current_local_path.is_dir():
            current_url_path.mkdir(parents=True, exist_ok=True)
            copy_tree(current_local_path.as_posix(), current_url_path.as_posix(), preserve_symlinks=True)


def dump_html_to_dir(static_dir_path, template):
    pathlib.Path(static_dir_path / template.template.name).write_bytes(template.body)


def upload_to_supervisely(static_dir_path):
    api: sly.Api = sly.Api.from_env()

    team_id, task_id = int(os.getenv('context.teamId')), int(os.getenv('TASK_ID', 0000))
    remote_dir = pathlib.Path('/', 'offline-sessions', str(task_id), 'app-template')

    res_remote_dir: str = api.file.upload_directory(team_id=team_id,
                                                    local_dir=static_dir_path.as_posix(),
                                                    remote_dir=remote_dir.as_posix())

    if os.getenv('TASK_ID') is not None:
        api.task.update_meta(id=task_id, data={'templateRootDirectory': res_remote_dir})

    sly.logger.info(f'App files stored in {res_remote_dir} for offline usage')


def dump_files_to_supervisely(app: FastAPI, template_response):
    try:
        if os.getenv('_SUPERVISELY_OFFLINE_FILES_UPLOADED', 'False') == 'True' and \
                template_response.context.get('request') is not None:
            return
        os.environ['_SUPERVISELY_OFFLINE_FILES_UPLOADED'] = 'True'
        sly.logger.info(f'Saving app files for offline usage')

        app_template_path = pathlib.Path(tempfile.mkdtemp())
        app_static_paths = get_static_paths_by_mounted_object(mount=app)
        dump_statics_to_dir(static_dir_path=app_template_path, static_paths=app_static_paths)
        dump_html_to_dir(static_dir_path=app_template_path, template=template_response)

        upload_to_supervisely(static_dir_path=app_template_path)

        shutil.rmtree(app_template_path.as_posix())

    except Exception as ex:
        traceback.print_exc()
        sly.logger.warning(f'Cannot dump files for offline usage, reason: {ex}')
        os.environ['_SUPERVISELY_OFFLINE_FILES_UPLOADED'] = 'False'


def available_after_shutdown(app: FastAPI):
    def func_layer_wrapper(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            template_response = f(*args, **kwargs)
            try:
                threading.Thread(
                    target=functools.partial(dump_files_to_supervisely, app, template_response),
                    daemon=False
                ).start()

                return template_response
            except Exception as ex:
                traceback.print_exc()
                sly.logger.warning(f'Cannot dump files for offline usage, reason: {ex}')

        return wrapper

    return func_layer_wrapper



from supervisely.geometry.geometry import Geometry
from supervisely.geometry.constants import (
    EXTERIOR,
    INTERIOR,
    POINTS,
    LABELER_LOGIN,
    UPDATED_AT,
    CREATED_AT,
    ID,
    CLASS_ID,
)


class ClosedSurfaceMesh(Geometry):
    @staticmethod
    def geometry_name():
        return "closed_surface_mesh"

    def draw(self, bitmap, color, thickness=1, config=None):
        raise NotImplementedError('Method "draw" is unavailable for this geometry')

    def draw_contour(self, bitmap, color, thickness=1, config=None):
        raise NotImplementedError(
            'Method "draw_contour" is unavailable for this geometry'
        )

    def convert(self, new_geometry, contour_radius=0, approx_epsilon=None):
        raise NotImplementedError('Method "convert" is unavailable for this geometry')

    def to_json(self):
        res = {}
        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, data):
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)

        return cls(
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )



# coding: utf-8
"""list all user roles that are available on private supervisely instance"""

# docs
from __future__ import annotations
from typing import List, Dict, Optional

from enum import IntEnum
from supervisely.api.module_api import ApiField, ModuleApiBase


class RoleApi(ModuleApiBase):
    """
    API for working with Roles. :class:`RoleApi<RoleApi>` object is immutable.

    :param api: API connection to the server
    :type api: Api
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # You can connect to API directly
        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Or you can use API from environment
        os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()

        roles = api.role.get_list() # api usage example
    """
    class DefaultRole(IntEnum):
        ADMIN =     1
        DEVELOPER = 2
        ANNOTATOR = 3
        VIEWER =    4

    @staticmethod
    def info_sequence():
        """
        NamedTuple RoleInfo information about Role.

        :Example:

         .. code-block:: python

            RoleInfo(id=71,
                     role='manager',
                     created_at='2019-12-10T14:31:41.878Z',
                     updated_at='2019-12-10T14:31:41.878Z')
        """
        return [ApiField.ID,
                ApiField.ROLE,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **RoleInfo**.
        """
        return 'RoleInfo'

    def get_list(self, filters: Optional[List[Dict[str, str]]]=None) -> List:
        """
        List of all roles that are available on private Supervisely instance.

        :param filters: List of params to sort output Roles.
        :type filters: list
        :return: List of all roles with information. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`list`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            roles = api.role.get_list()
        """
        return self.get_list_all_pages('roles.list', {ApiField.FILTER: filters or []})



import SimpleITK as sitk

from supervisely.api.module_api import ApiField, RemoveableBulkModuleApi
from supervisely.api.volume.volume_annotation_api import VolumeAnnotationAPI
from supervisely.api.volume.volume_object_api import VolumeObjectApi
from supervisely.api.volume.volume_figure_api import VolumeFigureApi

# from supervisely.api.volume.video_frame_api import VolumeFrameAPI
from supervisely.api.volume.volume_tag_api import VolumeTagApi

from supervisely.io.fs import ensure_base_path

from supervisely.io.fs import (
    get_file_ext,
    get_file_name,
    get_bytes_hash,
)
from supervisely import volume
import supervisely.volume.nrrd_encoder as nrrd_encoder
from supervisely._utils import batched
from supervisely import logger
from supervisely.task.progress import Progress


class VolumeApi(RemoveableBulkModuleApi):
    def __init__(self, api):
        """
        :param api: Api class object
        """
        super().__init__(api)
        self.annotation = VolumeAnnotationAPI(api)
        self.object = VolumeObjectApi(api)
        # self.frame = VideoFrameAPI(api)
        self.figure = VolumeFigureApi(api)
        self.tag = VolumeTagApi(api)

    @staticmethod
    def info_sequence():
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.LINK,
            ApiField.HASH,
            ApiField.MIME,
            ApiField.EXT,
            ApiField.SIZEB3,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.META,
            ApiField.PATH_ORIGINAL,
            ApiField.FULL_STORAGE_URL,
            ApiField.TAGS,
            ApiField.TEAM_ID,
            ApiField.WORKSPACE_ID,
            ApiField.PROJECT_ID,
            ApiField.DATASET_ID,
            ApiField.FILE_META,
            ApiField.FIGURES_COUNT,
            ApiField.ANN_OBJECTS_COUNT,
            ApiField.PROCESSING_PATH,
        ]

    @staticmethod
    def info_tuple_name():
        return "VolumeInfo"

    def _convert_json_info(self, info: dict, skip_missing=True):
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        return res

    def get_list(self, dataset_id, filters=None, sort="id", sort_order="asc"):
        return self.get_list_all_pages(
            "volumes.list",
            {
                ApiField.DATASET_ID: dataset_id,
                ApiField.FILTER: filters or [],
                ApiField.SORT: sort,
                ApiField.SORT_ORDER: sort_order,
            },
        )

    def get_info_by_id(self, id):
        return self._get_info_by_id(id, "volumes.info")

    def upload_hash(self, dataset_id, name, hash, meta=None):
        metas = None if meta is None else [meta]
        return self.upload_hashes(dataset_id, [name], [hash], metas=metas)[0]

    def upload_hashes(self, dataset_id, names, hashes, progress_cb=None, metas=None):
        return self._upload_bulk_add(
            lambda item: (ApiField.HASH, item),
            dataset_id,
            names,
            hashes,
            progress_cb,
            metas=metas,
        )

    def _upload_bulk_add(
        self, func_item_to_kv, dataset_id, names, items, progress_cb=None, metas=None
    ):
        results = []

        if len(names) == 0:
            return results
        if len(names) != len(items):
            raise RuntimeError(
                'Can not match "names" and "items" lists, len(names) != len(items)'
            )

        if metas is None:
            metas = [{}] * len(names)
        else:
            if len(names) != len(metas):
                raise RuntimeError(
                    'Can not match "names" and "metas" len(names) != len(metas)'
                )

        for batch in batched(list(zip(names, items, metas))):
            volumes = []
            for name, item, meta in batch:
                item_tuple = func_item_to_kv(item)
                image_data = {ApiField.NAME: name, item_tuple[0]: item_tuple[1]}
                if len(meta) != 0 and type(meta) == dict:
                    image_data[ApiField.META] = meta
                volumes.append(image_data)

            response = self._api.post(
                "volumes.bulk.add",
                {ApiField.DATASET_ID: dataset_id, ApiField.VOLUMES: volumes},
            )
            if progress_cb is not None:
                progress_cb(len(volumes))

            for info_json in response.json():
                results.append(self._convert_json_info(info_json))
        return results

    def upload_np(
        self, dataset_id, name, np_data, meta, progress_cb=None, batch_size=30
    ):
        ext = get_file_ext(name)
        if ext != ".nrrd":
            raise ValueError(
                "Name has to be with .nrrd extension, for example: my_volume.nrrd"
            )
        from timeit import default_timer as timer

        logger.info("Start volume compression before upload...")
        start = timer()
        volume_bytes = volume.encode(np_data, meta)
        logger.info(f"Volume has been compressed in {timer() - start} seconds")

        logger.info(f"Start uploading bytes of 3d volume ...")
        start = timer()
        volume_hash = get_bytes_hash(volume_bytes)
        self._api.image._upload_data_bulk(lambda v: v, [(volume_bytes, volume_hash)])
        logger.info(
            f"3d Volume bytes has been sucessfully uploaded in {timer() - start} seconds"
        )
        volume_info = self.upload_hash(dataset_id, name, volume_hash, meta)
        if progress_cb is not None:
            progress_cb(1)  # upload volume

        # slice all directions
        # http://insightsoftwareconsortium.github.io/SimpleITK-Notebooks/Python_html/03_Image_Details.html#Conversion-between-numpy-and-SimpleITK
        # x = 1 - sagittal
        # y = 1 - coronal
        # z = 1 - axial

        for (plane, dimension) in zip(["sagittal", "coronal", "axial"], np_data.shape):
            for batch in batched(list(range(dimension))):
                slices = []
                slices_bytes = []
                slices_hashes = []
                try:
                    for i in batch:
                        normal = {"x": 0, "y": 0, "z": 0}

                        if plane == "sagittal":
                            pixel_data = np_data[i, :, :]
                            normal["x"] = 1
                        elif plane == "coronal":
                            pixel_data = np_data[:, i, :]
                            normal["y"] = 1
                        elif plane == "axial":
                            pixel_data = np_data[:, :, i]
                            normal["z"] = 1
                        else:
                            raise ValueError(f"Unknown plane {plane}")

                        img_bytes = nrrd_encoder.encode(
                            pixel_data, header={"encoding": "gzip"}, compression_level=1
                        )

                        img_hash = get_bytes_hash(img_bytes)
                        slices_bytes.append(img_bytes)
                        slices_hashes.append(img_hash)
                        slices.append(
                            {
                                "hash": img_hash,
                                "sliceIndex": i,
                                "normal": normal,
                            }
                        )

                    if len(slices) > 0:
                        self._api.image._upload_data_bulk(
                            lambda v: v,
                            zip(slices_bytes, slices_hashes),
                        )
                        self._upload_slices_bulk(volume_info.id, slices, progress_cb)

                except Exception as e:
                    exc_str = str(e)
                    logger.warn(
                        "File skipped due to error: {}".format(exc_str),
                        exc_info=True,
                        extra={
                            "exc_str": exc_str,
                            "file_path": name,
                        },
                    )
        return volume_info

    def upload_dicom_serie_paths(self, dataset_id, name, paths, log_progress=True, anonymize=True):
        volume_np, volume_meta = volume.read_dicom_serie_volume_np(paths, anonymize=anonymize)
        progress_cb = None
        if log_progress is True:
            progress_cb = Progress(
                f"Upload volume {name}", sum(volume_np.shape)
            ).iters_done_report
        res = self.upload_np(dataset_id, name, volume_np, volume_meta, progress_cb)
        return self.get_info_by_name(dataset_id, name)

    def _upload_slices_bulk(self, volume_id, items, progress_cb=None):
        results = []
        if len(items) == 0:
            return results

        for batch in batched(items):
            response = self._api.post(
                "volumes.slices.bulk.add",
                {ApiField.VOLUME_ID: volume_id, ApiField.VOLUME_SLICES: batch},
            )
            if progress_cb is not None:
                progress_cb(len(batch))
            results.extend(response.json())
        return results

    def upload_nrrd_serie_path(self, dataset_id, name, path, log_progress=True):
        volume_np, volume_meta = volume.read_nrrd_serie_volume_np(path)
        progress_cb = None
        if log_progress is True:
            progress_cb = Progress(
                f"Upload volume {name}", sum(volume_np.shape)
            ).iters_done_report
        res = self.upload_np(dataset_id, name, volume_np, volume_meta, progress_cb)
        return self.get_info_by_name(dataset_id, name)

    def _download(self, id, is_stream=False):
        response = self._api.post(
            "volumes.download", {ApiField.ID: id}, stream=is_stream
        )
        return response

    def download_path(self, id, path, progress_cb=None):
        response = self._download(id, is_stream=True)
        ensure_base_path(path)

        with open(path, "wb") as fd:
            mb1 = 1024 * 1024
            for chunk in response.iter_content(chunk_size=mb1):
                fd.write(chunk)

                if progress_cb is not None:
                    progress_cb(len(chunk))

    def upload_nrrd_series_paths(
        self,
        dataset_id,
        names,
        paths,
        progress_cb=None,
    ):
        volume_infos = []
        for name, path in zip(names, paths):
            info = self.upload_nrrd_serie_path(
                dataset_id, name, path, log_progress=True
            )
            volume_infos.append(info)
            if progress_cb is not None:
                progress_cb(1)
        return volume_infos



from supervisely.app import DataJson
from supervisely.app.widgets import Widget

INFO = "info"
WARNING = "warning"
ERROR = "error"


class DoneLabel(Widget):
    def __init__(
        self,
        text: str = None,
        widget_id: str = None,
    ):
        self._text = text
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {"text": self._text}

    def get_json_state(self):
        return None

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        self._text = value
        DataJson()[self.widget_id]['text'] = self._text



import copy
import functools
from enum import Enum
from typing import Literal

import fastapi
from varname import varname

from supervisely.app import DataJson
from supervisely.app.fastapi import run_sync
from supervisely.app.widgets import Widget


class ElementButton(Widget):
    class Routes:
        BUTTON_CLICKED = 'button_clicked_cb'

    def __init__(self,
                 text: str = 'Element Button',
                 button_type: Literal["primary", "info", "warning", "danger", "success"] = "primary",
                 button_size: Literal["mini", "small", "large"] = None,
                 plain: bool = False,
                 widget_id: str = None):
        self._widget_routes = {}

        self._text = text
        self._button_type = button_type
        self._plain = plain
        self._button_size = button_size

        self._loading = False
        self._disabled = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            'text': self._text,
            'button_type': self._button_type,
            'plain': self._plain,
            'button_size': self._button_size,

            'loading': self._loading,
            'disabled': self._disabled
        }

    def get_json_state(self):
        return None

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        self._text = value
        DataJson()[self.widget_id]['text'] = self._text

    @property
    def loading(self):
        return self._loading

    @loading.setter
    def loading(self, value):
        self._loading = value
        DataJson()[self.widget_id]['loading'] = self._loading

    @property
    def disabled(self):
        return self._disabled

    @disabled.setter
    def disabled(self, value):
        self._disabled = value
        DataJson()[self.widget_id]['disabled'] = self._disabled




# coding: utf-8

# docs
from typing import Optional
from uuid import UUID

from supervisely._utils import take_with_default
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_figure import VideoFigure
from supervisely.pointcloud_annotation.pointcloud_object_collection import PointcloudObjectCollection


class PointcloudFigure(VideoFigure):
    """
    Class for creating and using PointcloudFigure objects for videos
    """
    def __init__(self, parent_object, geometry, frame_index: Optional[int]=None, key: Optional[UUID]=None,
                 class_id: Optional[int]=None, labeler_login: Optional[str]=None, updated_at: Optional[str]=None, created_at: Optional[str]=None):
        """
        :param parent_object:  VideoObject class object
        :param geometry: Geometry class object (cuboid_3d only)
        :param frame_index: int
        :param key: uuid class object
        """
        super().__init__(parent_object, geometry, frame_index, key, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)
        #@TODO: validate geometry - allowed: only cuboid_3d + point_cloud

    def validate_bounds(self, img_size, _auto_correct=False):
        raise NotImplementedError()
